{ 4A20 low level I/O procedures }
{$IFNDEF main}
uses DOS;                                      { needed for the interrupt vector procedures }
{$ENDIF}

{$I hprint}
{ billions and billions of constants }
const
RateClock       = 4194300;                  { all 4A20s }

{ define 4A20 register locations }
A20Base         = $0220;                       { default 4A20 location (determined by U18) }

ADCDataRegLow   : word = A20Base + $00;
ADCDataRegHigh  : word = A20Base + $01;

ConvStatusReg   : word = A20Base + $02;
ADStartReg      : word = A20BAse + $03;
GCtable         : word = A20BAse + $04;
GCTableIncLoc   : word = A20BAse + $05;
ADModeReg       : word = A20Base + $06;
GCTableClrLoc   : word = A20BAse + $07;

SetupReg        : word = A20Base + $07;

ADCounter0      : word = A20Base + $08;
ADCounter1      : word = A20Base + $09;
ADCounter2      : word = A20Base + $0A;
ADCounterCont   : word = A20Base + $0B;                 { Normally not used - generated }

{******* data and masks that apply to 4A20 registers ********}

{ masks and data that apply to ConvStatusReg }

ADEOCMask       : byte = $01;                  { to check end of conversion bit }
PolledStart     : byte = $01;                  { to start conversion in polled mode }

{ masks and data that apply to ADStartReg }

ADStartMask     : byte = $01;                  { to check go bit }
ADStart         : byte = $01;                  { to set go bit }
ADStop          : byte = $00;                  { to clear go bit }

{ masks and data that apply to ADModeReg }

PolledMode      : byte = $00;                  { to select polled mode of operation }
InterruptMode   : byte = $01;                  { to select interrupt driven mode of operation }
DMAMode0        : byte = $02;                  { to select DMA driven - single conversion per EXTSC mode }
DMAMode1        : byte = $03;                  { to select DMA driven - block conversion per EXTSC mode }

{ masks and data that apply to GC table bytes }

ChannelMask     : byte = $3F;                  { to avoid messing with gain bits ... }
ADGain1         = $00;                         { select gain of 1 }
ADGain10        = $40;                         { select gain of 10 }
ADGain100       = $80;                         { select gain of 100 }

GCTableLength   : word = $1000;                 { length of channel and gain table is 2048 bytes }

{ masks and data that apply to SetupReg }

PowerDwn            : byte = $02;                  { to turn power off }
WordMode            : byte = $01;                  { 16 bit mode }

{*********************** 8254 COUNTER TIMER ROUTINES ***********************}

{$I 8254}

{******************* 8259 INTERRUPT CONTROLLER ROUTINES ********************}

{$I 8259}

{************************** 8237 DMA RELATED ROUTINES **********************}

{$I 8237 }

{********************* LOW LEVEL 4A20 SPECIFIC PROCEDURES ******************}

procedure WaitABit(loopCount : integer);       { sloppy wait for short non-critical delays }
var
loops : integer;
dummy : byte;
begin
  for loops := 1 to loopCount do
  begin
    { I/O reads are slow on most fast machines }
    dummy := port[$20];
  end
  {nothing} ;
end;

procedure StopConvertClock;
begin
  port [ADStartReg] := ADStop;
end;

procedure StartConvertClock;
begin
  port [ADStartReg] := ADStart;
end;

procedure StartPolledConversion;
begin
  port[ConvStatusReg] := PolledStart;
end;

procedure WriteSetup(inum,dnum,smode : byte);
const

irqs :array[0..15] of byte =($FF,$FF,$00,$FF,$FF,$01,$02,$03,
                        {IRQ:  0   1   2   3   4   5   6   7}
                             $FF,$00,$04,$05,$06,$FF,$FF,$07);
                        {IRQ:  8   9  10  11  12  13  14  15}
begin
{ First convert irq number from caller to proper select code }
  inum := inum and $0F;
  inum := irqs[inum];
  if inum = $FF then writeln('Invalid 4A20 interrupt selected');
  SelectDMAChannel(dnum); { this selects proper I/O port addresses }
                          { for the first or second DMA controller }
                          { plus selects 16 bit DMA mode if necessary  }
                          { doesn't actually touch DMA controller... }
  port[SetupReg] := smode or (inum shl 2) or (dnum shl 5);
end;

procedure ClrGCTablePtr;
var data : byte;
begin
  data := port[GCTableClrLoc]; { read it}
  port[GCTableClrLoc] := data and 3; { clear the table pointer without changing setup }
end;

procedure IncGCTablePtr;
begin
  port[GCTableIncLoc] := 0;                    { increment the table pointer (data written is unimportant) }
end;

procedure LoadScanLengthCounter(scanlength : word);
begin                                          { because 8254 type counters need a clock }
  StopConvertClock;                            { transition to load the counter, we set }
  SetCounterMode(ADCounter0,SoftwStrobe);      { the rate generator into one shot mode. }
  SetCounterMode(ADCounter1,RateGenerator);    { setup proper scan counter mode }
  StartConvertClock;                           { re-enable clock once were in one shot mode }
  LoadCounter(ADCounter0,1);                   { generate a pulse to ensure output polarity }
  LoadCounter(ADCounter1,scanlength);          { preload the scanlength counter }
  LoadCounter(ADCounter0,1);                   { first pulse loads scanlength }
  WaitABit(10);                                { wait a little to ensure one shot done on fast machines... }
  LoadCounter(ADCounter0,1);                   { second pulse to syncronize scanlength counter }
  WaitABit(10);                                { wait a little to ensure one shot done on fast machines... }
  ClrGCTablePtr;                               { with cleared table pointer }
  StopConvertClock;                            { prepare for start }
end;

procedure LoadRateCounter(ratedivisor : word);
begin
  SetCounterMode(ADCounter0,RateGenerator);    { set counter to rate generator mode }
  LoadCounter(ADCounter0,ratedivisor);         { set rate }
end;

procedure SetMain4A20Mode(mode : byte);
begin
  port[ADModeReg]:= mode;                      { set operational mode }
end;

procedure IdleScanLengthCounter;               { to keep output of scan counter high, }
begin                                          { the do - nothing state }
  SetCounterMode(ADCounter1,HardwStrobe);      { gate is tied high ... therefore no trigger }
  LoadCounter(ADCounter1,1);                   { minimum pulse width just in case something happens }
end;

procedure IdleRateCounter;                     { to keep output of rate counter high, }
begin                                          { for polled mode, the do - nothing state }
  SetCounterMode(ADCounter0,SoftwStrobe);      { gate is indeterminate }
  LoadCounter(ADCounter0,1);                   { generate 2 usec pulse if gate high ... }
end;

procedure Clr4A20IIRQ;
var
nothing : byte;
begin
  nothing := port[ADCDataRegLow];              { read the ADC to clear the IRQ FF }
end; { Clr4A20IIRQ }

procedure Reset4A20;                           { do full table ptr clear - may take a while ... }
begin                                          { note that this routine destroys any information in the ADModeReg }
  ClrGCTablePtr;                               { clear table pointer }
  port[SetupReg] := 0;
  StopConvertClock;
  IdleScanLengthCounter;
  IdleRateCounter;                             { keep timers out of trouble }
  SetMain4A20Mode(PolledMode);                 { must be in polled mode to do reset }
  Clr4A20IIrq;                                 { clear the IRQ FF, if set }
end; { Reset4A20 }

procedure SetChannelAndGain(channel: byte; gain: byte);
begin
  port[GCTable] := (channel and ChannelMask) or gain;
end;

{ FillGCTableWithSequentialChannels must be followed by setting the }
{ 4A20 mode and (re)writing the setup register  }

procedure FillGCTableWithSequentialChannels(gain: byte);
var
index : word;
begin
  SetMain4A20Mode(PolledMode);                 { prevent anyone else incrementing pointer }
  ClrGCTablePtr;                               { note that we destroy any existing setup information }
  index :=0;
  while index < GCTableLength do               { from 0 to GCTableLength -1 }
  begin
    SetChannelAndGain((index and ChannelMask),gain);   { write sequential channel numbers into table }
    IncGCTablePtr;                             { increment the table pointer }
    index := index + 1;
  end; { while }
end; { FillGCTableWithSequentialChannels }

{ FillGCTableWithOffsetChannels must be followed by setting the     }
{ 4A20 mode and (re)writing the setup register This routine loads   }
{ the channel and gain table with  sequential channels up to the    }
{ chans number and applied a 1 count offset to compensate for the   }
{ A-D pipeline.                                                     }
{ If chans is 8, the sequence is 1 2 3 4 5 6 7 0 1 2 3 . . .        }

procedure FillGCTableWithOffsetChannels(chans: byte;gain: byte);
var
index : word;
chan : byte;
begin
  SetMain4A20Mode(PolledMode);                 { prevent anyone else incrementing pointer }
  ClrGCTablePtr;                               { note that we destroy any existing setup information }
  index :=0;
  chan  :=1;
  while index < GCTableLength do               { fill whole table... }
  begin
    SetChannelAndGain(chan and channelMask,gain);   { write sequential channel numbers into table }
    IncGCTablePtr;                             { increment the table pointer }
    index := index+1;
    chan := chan+1;
    if chan = chans then chan := 0;
  end; { while }
end; { FillGCTableWithOffsetChannels }

procedure FillGCTable(start: integer; length: integer; channel: byte; gain : byte);
var
index : word;
begin
  SetMain4A20Mode(PolledMode);                 { prevent anyone else incrementing pointer }
  ClrGCTablePtr;                               { note that we destroy any existing mode information }
  index :=0;
  while index < start do
  begin
    IncGCTablePtr;                             { get to the first }
    index := index + 1;
  end;
  index := start;                              { insecurity ...}
  while index <= (start + length) do
  begin
    SetChannelAndGain(channel,gain);           { write channel and gain entry in table }
    IncGCTablePtr;                             { increment the table pointer }
    index := index + 1;
  end; { while }
end; { FillGCTable }

{ the following two procedures are for use by table RAM testing only }

procedure WriteGCTableWord(wdata : word);
begin
  port[GCTable] := lo(wdata);                { write low byte }
  IncGCTablePtr;                             { increment the table pointer }
  port[GCTable] := hi(wdata);                { write high byte }
  IncGCTablePtr;                             { increment the table pointer }
end;


procedure ReadGCTableWord(var rdata : word);
begin
  rdata := port[GCTable];                      { read low byte }
  IncGCTablePtr;                               { increment the table pointer }
  rdata := swap(rdata);                        { prepare for high byte }
  rdata := rdata or port[GCTable];             { or in high byte }
  rdata := swap(rdata);                        { swippity swappity bippity boppity }
  IncGCTablePtr;                               { increment the table pointer }
end;


Function CheckGCTable : boolean;               { check the GCtable RAM and logic }
var                                            { this is useful as a check of 4A20 presence }
index : word;
checkdata : word;
begin
  CheckGCTable := true;                        { good unless we find otherwise }
  SetMain4A20Mode(PolledMode);                  { prevent anyone else incrementing pointer }
  { do address check }
  ClrGCTablePtr;                               { clear the table pointer }
  index :=0;                                   { likewise our index }
  while index < GCTableLength do               { from 0 to GCTableLength -2 }
  begin
    WriteGCTableWord(index);                   { load table with sequential 16 bit numbers }
    index := index + 2;
  end; { while index }
  ClrGCTablePtr;                               { clear the table pointer }
  index :=0;                                   { likewise our index }
  while index < GCTableLength do               { from 0 to GCTableLength -2 }
  begin
    ReadGCTableWord(checkdata);                { read and check the words }
    if checkData <> index then
    begin
      CheckGCTable := false;
      writeln('Error checking table RAM');
      write('Expected: ');
      HexPrint(index,2);
      write(' but found ');
      HexPrint(checkData,2);
      writeln;
    end;
    index := index + 2;
  end; { while index }
  { complemented address check }
  ClrGCTablePtr;                               { clear the table pointer }
  index :=0;                                   { likewise our index }
  while index < GCTableLength do               { from 0 to GCTableLength -2 }
  begin
    WriteGCTableWord(index xor $FFFF);         { load table with complements of sequential 16 bit numbers }
    index := index + 2;
  end; { while index }
  ClrGCTablePtr;                               { clear the table pointer }
  index :=0;                                   { likewise our index }
  while index < GCTableLength do               { from 0 to GCTableLength -2 }
  begin
    ReadGCTableWord(checkData);                { read and check the words }
    if checkData <> (index xor $FFFF) then CheckGCTable := false;
    index := index + 2;
  end; { while index }
end; { CheckGCTable }


{ PolledConverterRead is selects the desired channel and gain, }
{ waits a settling time and then does the conversion           }


function PolledConverterRead(channel : byte; gain : byte): longint;
var

adata,longdata : longint;

begin
  SetChannelAndGain(channel,gain);                { select channel and gain }
  WaitABit(50);                                   { wait 20 usec min. }
  StartPolledConversion;
  while (port[ConvStatusReg] and ADEOCMask) = 0 do; { throw away first conversion }
  StartPolledConversion;                          { to avoid A-D pipelining     }
  while (port[ConvStatusReg] and ADEOCMask) = 0 do; { nothing }
  adata := portw[ADCDataReglow];                  { read converted data }
  if adata >32767 then adata := ((adata and $7FFF) -32768);
  longdata := longint(adata);
  PolledConverterRead := longdata;
end; { PolledConverterRead }

function PolledConverterReadWithAverage(channel : byte; gain : byte;times : longint): longint;
var
adata,longdata,average : longint;

begin
  SetChannelAndGain(channel,gain);             { select channel and gain }
  WaitABit(50);                                { wait 20 usec min. }
  StartPolledConversion;
  while port[ConvStatusReg] and ADEOCMask = 0 do;{ throw away first conversion }
  longdata := 0;
  for average := 1 to times do
  begin
    StartPolledConversion;                         { to avoid A-D pipelining     }
    while (port[ConvStatusReg] and ADEOCMask) = 0 do; { nothing }
    adata := longint(portw[ADCDataReglow]);                  { read converted data }
    if adata >32767 then adata := ((adata and $7FFF) -32768);
    longdata := longdata + adata;
  end;
  longdata := longdata div times;
  PolledConverterReadWithAverage := longdata;
end; { PolledConverterRead }

{ IntTableConverterRead assumes that the channel and gain were set up }
{ from the table and that it was called in response to an EOC interrupt }

function IntTableConverterRead : word;
var
adata : word;
begin
  IncGCTablePtr;                               { increment the table pointer }
  adata := portw[ADCDataReglow];               { read converted data }
  IntTableConverterRead:= adata;
end; { IntTableConverterRead }

{ InterruptConverterRead assumes that the channel and gain were set up on }
{ the previous InterruptConverterRead, reads the converter, and sets up }
{ channel and gain for the next InterruptConverterRead }

function InterruptConverterRead(nextchannel : byte; nextgain : byte) : word;
var
adata : word;
begin
  SetChannelAndGain(nextChannel,nextGain);     { select channel and gain for next reading }
  adata := portw[ADCDataReglow];               { read converted data }
  InterruptConverterRead:= adata;
end; { InterruptConverterRead}

